from cmath import isnan
import numpy as np
import random
import hashlib
import math

def get_state_id(state):
    return ','.join([str(x) for x in sorted(state)])

class Point():
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def id(self):
        return ','.join([str(int(x)) for x in self.x])

class Influence():
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.original_dof = set()
        self.dof = set()
        for i in range(0, len(a.x)):
            if a.x[i] != b.x[i]:
                self.original_dof.add(i)
                self.dof.add(i)

    def coherent(self):
        return self.a.y == self.b.y

    def id(self):
        return ','.join(sorted([self.a.id(), self.b.id()]))

def encode(v):
    byte_values = []
    for i in range(0, math.ceil(len(v) / 8)):
        x = 0
        for j in range(0, 8):
            index = i * 8 + j
            if index >= len(v):
                continue
            x <<= 1
            x |= int(v[index])
        byte_values.append(x)
    return bytearray(byte_values)

def decode(x, N):
    index = 0
    output = np.zeros((N))
    while x > 0 and index < N:
        output[index] = x & 0b1
        x >>= 1
        index += 1
    return output

def sha(v):
    x = encode(v)
    m = hashlib.sha256()
    m.update(x)
    result = m.digest()
    return result[0] & 0b1

def hamming_distance(a, b):
    return np.sum(np.logical_xor(a.x, b.x))

def random_x(N):
    x = np.zeros((N))
    for i in range(0, N):
        x[i] = random.randint(0, 1)
    return x

def xor(x):
    # return sum(x) % 2
    half = int(len(x) / 2)
    return sum(x[:half]) % 2

def create_dof_map(influences):
    dof_map = {}
    for influence in influences:
        for i in influence.dof:
            if not i in dof_map:
                dof_map[i] = []
            dof_map[i].append(influence)
    return dof_map

def flip(influences, i):
    for influence in influences:
        if i in influence.dof:
            influence.a.y = int(influence.a.y) ^ 1

def remove_dof(dof_map, i, flip = False):
    for influence in dof_map[i]:
        influence.dof.remove(i)
        if flip:
            influence.a.y = int(influence.a.y) ^ 1
        # if len(influence.dof) == 0 and not influence.coherent():
        #     raise Exception('Invalid')
    del dof_map[i]

def solve(dof_map, all_influences, all_samples):
    eliminated = True
    while eliminated:
        eliminated = False
        for influence in all_influences:
            if len(influence.dof) == 1:
                i = next(iter(influence.dof))
                if influence.coherent:
                    remove_dof(dof_map, i)
                    eliminated = True
                else:
                    print('Forced', i)
                    remove_dof(dof_map, i, True)
                    eliminated = True
                
    lowest_dof = None
    for influence in all_influences:
        if not influence.coherent and len(influence.dof) > 1:
            if lowest_dof is None or len(influence.dof) < len(lowest_dof.dof):
                lowest_dof = influence

    flip = None
    highest_score = -1

    for i in lowest_dof.dof:
        per_point_scores = {}
        i_influences = dof_map[i]
        left = 0
        right = 0
        for influence in i_influences:
            if not influence.a in per_point_scores:
                per_point_scores[influence.a] = [0, 0]
            if not influence.b in per_point_scores:
                per_point_scores[influence.b] = [0, 0]
            if influence.coherent:
                per_point_scores[influence.a][0] += 1
                per_point_scores[influence.b][0] += 1
                left += 1
            else:
                per_point_scores[influence.a][1] += 1
                per_point_scores[influence.b][1] += 1
                right += 1
        print(i, left / (left + right))
        num = 0
        denom = 0
        for _, score in per_point_scores.items():
            if score[0] == score[1]:
                continue
            print(i, score)
            num += score[1] / (score[0] + score[1])
            denom += 1
        score = num / denom if denom > 0 else 0
        print(score)
    
    return None


# 1st row (n+1 choose k+1) * (1-(k mod 2))
# psuedopascal to compute the follow-on rows
# assuming solvability, we want to maximize the probability that our current state and our state with
# a particular single flip are one order apart in the correct direction



# 2, 0
# 2, 2, 0
# 2, 4, 2, 0
# 2, 6, 6, 2, 0
# 2, 8,12, 8, 2, 0
# 2,10,20,20,10, 2, 0

# 3,-9,19,-33,51,-73,99
# 3,-6,10,-14,18,-22,26
# 3,-3, 4, -4, 4, -4, 4
# 3, 0, 1,  0, 0,  0, 0
# 3, 3, 1,  1, 0,  0, 0
# 3, 6, 4,  2, 1,  0, 0
# 3, 9,10,  6, 3,  1, 0

#       4, 0, 4, 0
#     4, 4, 4, 4, 0
#   4, 8, 8, 8, 4, 0
# 4,12,16,16,12, 4, 0

#       5, 0,10, 0, 1
#     5, 5,10,10, 1, 1
#   5,
# 5,



# 3
#
# @1 [1, 2, 1]
# @2 [2, 2, 0]
# @3 [3, 0, 1]

# 5  [5, 10, 10, 5, 1] (5 choose 1, 5 choose 2, ...)
#
# @1 [1, 4, 6,  4, 1], [4, 6,  4, 1, 0] - 16, 15 - binomial (4 choose 0, 4 choose 1, 4 choose 2),
# @2 [2, 6, 6,  2, 0], [3, 4,  4, 3, 1] - 16, 15 - (4 choose 1) + (2 choose -1) - (2 choose 1)
# @3 [3, 6, 4,  2, 1], [2, 4,  6, 3, 0] - 16, 15 - (4 choose 2) + (2 choose -2) - (2 choose 2) + (2 choose -1) - (2 choose 1)
# @4 [4, 4, 4,  4, 0], [1, 6,  6, 1, 1] - 16, 15 - 
# @5 [5, 0, 10, 0, 1], [0, 10, 0, 5, 0] - 16, 15 - 

# @0 [0.0, 0.0, 0.0, 0.0, 0.0]
# @1 [0.2, 0.4, 0.6, 0.8, 1.0]
# @2 [0.4, 0.6, 0.6, 0.4, 0.0]
# @3 [0.6, 0.6, 0.4, 0.4, 1.0]
# @4 [0.8, 0.4, 0.4, 0.8, 0.0]
# @5 [1.0, 0.0, 1.0, 0.0, 1.0]

# 6
#
# @1 [1, 5, 10, 10, 5, 1]
# @2 [2, 8, 12, 8,  2, 0]
# @3 [3, 9, 10, 6,  3, 1]
# @4 [4, 8, 8,  8,  4, 0]
# @5 [5, 5, 10, 10, 1, 1]
# @6 [6, 0, 20, 0,  6, 0]

# last row, 1 if odd, 0 if even
# second to last, subtract 2 on odds, add 2 on evens

def compute_pseudopascal(N):
    dist = np.zeros((N, N))
    for j in range(0, N):
        dist[0][j] = math.comb(N - 1, j)
        dist[-1][j] = math.comb(N, j + 1) * (1 - (j % 2))
    for i in range(1, N):
        for j in range(0, i + 1):
            dist[i][j] = math.comb(i + 1, j + 1) * (1 - (j % 2))
        for k in range(i + 1, N):
            for j in reversed(range(0, k)):
                dist[i][j+1] = dist[i][j] + dist[i][j+1]
    return dist

def compute_distributions(N):
    dist = compute_pseudopascal(N)
    print(dist)
    for i in range(0, N):
        for j in range(0, N):
            denom = math.comb(N, j+1)
            dist[i][j] /= denom
    return dist

def confusion_probabilities(N, samples):
    sample_sizes = np.zeros(N)
    for i in range(0, len(samples)):
        a = samples[i]
        for j in range(0, len(samples)):
            b = samples[j]
            if i == j:
                continue
            distance = hamming_distance(a, b)
            sample_sizes[distance - 1] += 1

    confusion = np.zeros((N, N))
    dist = compute_pseudopascal(N)
    np.multiply(dist, 2 ** N, dist)
    # These are the probabilities that we might mix up any two orders given a particular sample size
    for i in range(0, N):
        for j in range(0, N):
            probability = 1.0
            for k in range(0, N):
                full_size = math.comb(N, k+1) * (2 ** N)
                sample_size = sample_sizes[k]
                num_unknowns = full_size - sample_size
                i_incoherent = dist[i][k]
                # Worst case, we sample only the coherent points, 
                i_min = max(i_incoherent - num_unknowns, 0) / full_size
                i_max = min(sample_size, i_incoherent) / full_size
                u = i_min + i_max / 2
                s = (i_max - i_min) / 2
                probability *= raised_cosine(dist[j][k] / full_size, u, s)
            confusion[i][j] = probability
    return confusion

def raised_cosine(x, u, s):
    if x < (u - s):
        return 0
    if x > (u + s):
        return 0
    return 1.0 / (2.0 * s) * (1 + math.cos(math.pi * (x - u) / s))

# Probability of getting k red balls after drawing n from a bag with m total balls and j red balls in it
# (n choose k) * p^k * (1-p)^(n-k)

# p/m chance of getting a red ball
# (1 - p/m) chance of not getting a red ball

# One way (p/m) * ((p-1)/(m-1)) * ((p-2)/(m-2)) 
# (1 - (p/m))

def p_bernoulli(n, k, m, j):
    probabilities = np.zeros((n + 1, n + 1))
    probabilities.fill(-1)
    # if n == k:
    #     return 1.0
    # if k > p:
    #     return 0.0
    stack = [(0,0)]
    while len(stack) > 0:
        (a, b) = stack.pop()
        if a + b == n:
            probabilities[a][b] = 1 if a == k else 0
        elif a > j:
            probabilities[a][b] = 0
        elif b > (m - j):
            probabilities[a][b] = 0
        else:
            p_left = probabilities[a + 1][b]
            p_right = probabilities[a][b + 1]
            if p_left >= 0 and p_right >= 0:
                p = (j - a) / (m - a - b)
                probabilities[a][b] = p_left * p + p_right * (1 - p)
            else:
                stack.append((a, b))
                if p_left < 0:
                    stack.append((a + 1, b))
                if p_right < 0:
                    stack.append((a, b + 1))
    return probabilities[0][0]
    
    # P = 1.0
    # p_k = 0
    # p_nk = 0
    # for i in range(1, k + 1):
    #     P *= (n + 1 - i) / i
    #     while P > 1.0 and p_k < k:
    #         P *= p
    #         p_k += 1
    #     while P > 1.0 and p_nk < (n - k):
    #         P *= (1 - p)
    #         p_nk += 1
    # while p_k < k:
    #     P *= p
    #     p_k += 1
    # while (p_nk < (n - k)):
    #     P *= (1 - p)
    #     p_nk += 1
    # return P

def average_index(x):
    total = 0
    for k in range(0, len(x)):
        total += k * x[k]
    return total / np.sum(x)

def compute_cumulative_probability(N, bases, p_n):
    # p_n = np.zeros(N)
    # p_n.fill(0.5)
    states = [[]]
    flips = set()
    for i in range(1, len(bases)):
        # (base, _) = bases[i]
        (_, flip) = bases[i]
        # p_forward = 0
        # p_backward = 0
        # for k in range(0, N - 1):
        #     p_forward += base[k + 1] * next_p[k]
        #     p_backward += base[k] * next_p[k + 1]
        if flip in flips:
            # p_n[flip] -= p_forward
            # p_n[flip] += p_backward
            flips.remove(flip)
        else:
            # p_n[flip] += p_forward
            # p_n[flip] -= p_backward
            flips.add(flip)
        states.append(flips.copy())
    # np.clip(p_n, 0, 1, p_n)
    # print('Contribution probabilities', p_n)

    min_p_n = np.min(p_n)
    max_p_n = np.max(p_n)


    p_k = np.zeros(N)
    for k in range(0, N):
        stack = [(k, len(bases) - 1)]
        probabilities = np.zeros((N, len(bases)))
        probabilities.fill(-1)
        while len(stack) > 0:
            (i, base_index) = stack.pop()
            (base, flip) = bases[base_index]
            if base_index == 0:
                probabilities[i, 0] = base[i]
            else:
                left = i - 1
                right = i + 1
                state = states[base_index - 1]
                p_flip = max(min(p_n[flip] + 0.5, 1.0), 0)
                if flip in state:
                    p_flip = 1 - p_flip
                p_left = probabilities[left, base_index - 1] if left >= 0 else 0
                p_right = probabilities[right, base_index - 1] if right < N else 0
                if p_left >= 0 and p_right >= 0:
                    probabilities[i, base_index] = base[i] * p_left * (1 - p_flip) + base[i] * p_right * p_flip
                else:
                    stack.append((i, base_index))
                    if p_left < 0:
                        stack.append((left, base_index - 1))
                    if p_right < 0:
                        stack.append((right, base_index - 1))
        p_k[k] = probabilities[k][-1]
    np.divide(p_k, np.sum(p_k), p_k)
    return p_k

# 8, 32,    2^5
# 10, 64,   2^6
# 12, 128,  2^7
# 14, 256,  2^8
# 16, 512,  2^9
# 18, 1024, 2^10
# 20, 2048, 2^11
# 22, 4096, 2^12
def main():
    N = 8
    sample_size = 16
    sample_ids = set()
    samples = []

    dist = compute_pseudopascal(N)
    print(dist)

    for i in range(0, sample_size):
        x = random_x(N)
        y = int(xor(x))
        p = Point(x, y)
        p_id = p.id()
        if p_id in sample_ids:
            continue
        sample_ids.add(p_id)
        samples.append(p)
    # confusion = confusion_probabilities(N, samples)
    # print(confusion)
    # return

    # for i in range(0, 2**N):
    #     x = decode(i, N)
    #     y = int(xor(x))
    #     samples.append(Point(x,y))

    base = np.zeros(N)
    current = np.zeros(N)
    cumulative_probability = np.ones(N)
    flip_likelihood = np.zeros(N)
    cumulative_deltas = np.zeros(N)
    direction = -1
    flips = set()
    bases = []
    last_flip = -1

    for _ in range(0, 2 ** N):
        lowest_err = -1
        use_flip = -1
        for flip in range(-1, N):
            coherent_distances = np.zeros((len(samples), N+1))
            incoherent_distances = np.zeros((len(samples), N+1))
            all_coherent = True
            for i in range(0, len(samples)):
                a = samples[i]
                for j in range(0, len(samples)):
                    b = samples[j]
                    distance = hamming_distance(a, b)
                    is_coherent = ((flip < 0 or a.x[flip] == b.x[flip]) and a.y == b.y) or ((flip >= 0 and a.x[flip] != b.x[flip]) and a.y != b.y)
                    if is_coherent:
                        coherent_distances[i][distance] += 1
                    else:
                        incoherent_distances[i][distance] += 1
                        all_coherent = False
            if all_coherent:
                print('Flip and halt', flip)
                return
            # print(coherent_distances, incoherent_distances)

            # print(coherent_distances, incoherent_distances)
            # est_incoherence = np.divide(incoherent_distances, np.add(coherent_distances, incoherent_distances))
            # print(est_incoherence)

            probability = np.ones(N)
            np.divide(probability, np.sum(probability), probability)
            components = []
            for i in range(0, len(samples)):
                for j in range(1, N + 1):
                    p_k = np.zeros(N)
                    # confusion = np.zeros((N, N))
                    n = coherent_distances[i][j] + incoherent_distances[i][j]
                    if n == 0:
                        continue
                    a = incoherent_distances[i][j]
                    t = math.comb(N, j)
                    # for k in range(0, N):
                    #     p = dist[k][j - 1]
                    #     a_ideal = round(p * n / t)
                    #     # base_prob = p_bernoulli(int(n), a_ideal, t, int(p))
                    #     for q in range(0, N):
                    #         u = dist[q][j - 1]
                    #         p_ratio = p / t
                    #         u_ratio = u / t
                    #         confusion[k][q] = p_bernoulli(int(n), a_ideal, t, int(u))
                    # np.divide(confusion, np.max(confusion, axis=0), confusion)

                    for k in range(0, N):
                        p = dist[k][j - 1]
                        a_ideal = round(p * n / t)
                        # How likely are we to correctly identify an ideal sample?
                        # for q in range(0, N):
                        p_ideal = p_bernoulli(int(n), a_ideal, t, int(p))
                        # P = math.comb(int(n), int(a)) * math.pow(p, int(a)) * math.pow(1 - p, int(n - a))
                        p_k[k] = p_bernoulli(int(n), int(a), t, int(p))# * (n / t)
                        # p_bernoulli(int(n), int(a), math.comb(N, j), int(p))
                        # probability *= P
                    components.append(p_k)
                    np.divide(p_k, np.sum(p_k), p_k)
                    np.multiply(probability, p_k, probability)
                    np.divide(probability, np.sum(probability), probability)

                    # p_cross_k is the probability that we correctly identified at k 
                    # plus the probabilities that we missidentify at q and it is actually k

                    # probability of drawing from sample k = p_bernoulli

                    # p_cross_k = np.zeros(N)
                    # for k in range(0, N):
                    #     for q in range(0, N):
                    #         p_cross_k[k] += p_k[q] * confusion[k][q]
                            # if k == q:
                            #     continue
                            # p_cross_k[k] += (1 - p_k[k]) * p_k[q] * confusion[k][q]
                            # p_cross_k[k] -= (1 - p_k[q]) * p_k[k] * confusion[q][k]

                            # if q == k:
                            #     continue
                            # p_cross_k[k] += (1 - p_k[k]) * p_k[q] * confusion[k][q]
                            # p_cross_k[k] -= (1 - p_k[k])
                            # p_cross_k[k] -= p_k[k] * (1 - confusion[k][k]) * confusion[q][k]


                    # for k in range(0, N):
                    #     P = p_k[k]
                    #     for m in range(0, N):
                    #         if m == k:
                    #             continue
                    #         if p_k[m] == 0:
                    #             continue
                    #         P /= p_k[m]
                    #     p_cross_k[k] = P
                    # min_value = np.min(p_cross_k)
                    # np.subtract(p_cross_k, min_value, p_cross_k)
                    # np.add(probability, p_cross_k, probability)
                    # total = np.sum(p_k)
                    # if total > 0:
                    #     np.divide(p_k, total, p_k)
                    #     np.multiply(p_k, probability, probability)
                    #     np.divide(probability, np.sum(probability), probability)
                        # print(probability)


            np.divide(probability, np.sum(probability), probability)
            if flip < 0:
                np.copyto(base, probability)
            else:
                np.copyto(current, probability)


                    # print(k, i, min_true_value, max_true_value)

                    # confidence = (coherent_distances[i] + incoherent_distances[i]) / math.comb(N, i) # probability that the sample is representative
                    # err += abs(est_incoherence[i] - known_incoherence_at_k[i-1]) * confidence
                    # denom += 1
                # print(flip, k, err)
                # err /= denom
                # if flip < 0:
                #     base[k] = probability
                # else:
                #     current[k] = probability
            
            if flip >= 0:
                if np.sum(current) == 0:
                    continue
                np.divide(current, np.sum(current), current)

                base_mean_index = average_index(base)
                base_variance = 0
                for i in range(0, N):
                    base_variance += base[i] * (base_mean_index - i) ** 2
                base_err = 0
                norm = np.zeros(N)
                for i in range(0, N):
                    norm[i] = 1 / (base_variance * math.sqrt(2 * math.pi)) * math.exp(-1 / 2 * ((i - base_mean_index) / base_variance) ** 2)
                np.divide(norm, np.sum(norm), norm)
                for i in range(0, N):
                    base_err += (base[i] - norm[i]) ** 2
                
                current_mean_index = average_index(current)
                current_variance = 0
                for i in range(0, N):
                    current_variance += current[i] * (current_mean_index - i) ** 2
                current_err = 0
                for i in range(0, N):
                    norm[i] = 1 / (current_variance * math.sqrt(2 * math.pi)) * math.exp(-1 / 2 * ((i - current_mean_index) / current_variance) ** 2)
                np.divide(norm, np.sum(norm), norm)
                for i in range(0, N):                   
                    current_err += (current[i] - norm[i]) ** 2

                delta = abs(1 - (base_mean_index - current_mean_index))
                print(flip, current)
                print('Mean', current_mean_index, base_mean_index)
                print('Variance', current_variance, base_variance)
                print('Err', current_err, base_err)
                score = current_variance

                # base_score = 0
                # for i in range(0, N):
                #     base_score += (base[round(base_mean_index)] - base[i]) ** 2

                # score = 0
                # for i in range(0, N):
                #     score += (current[round(current_mean_index)] - current[i]) ** 2
                # print('Score', score, base_score)

                # print(current)
                # temp = np.roll(cumulative_probability, -1)
                # temp[-1] = 1.0
                # np.multiply(current, temp, current)
                # np.divide(current, np.sum(current), current)
                # p_forward = 0
                # p_backward = 0
                # for i in range(1, N):
                #     p_forward += base[i] * current[i - 1]
                # for i in range(0, N - 1):
                #     p_backward += base[i] * current[i + 1]
                # scale = 0.01
                # if flip in flips:
                #     flip_likelihood[flip] += scale * p_backward
                #     flip_likelihood[flip] -= scale * p_forward
                # else:
                #     flip_likelihood[flip] -= scale * p_backward
                #     flip_likelihood[flip] += scale * p_forward                    
                # delta = p_forward - p_backward
                # print(flip, current, p_forward, p_backward)
                # base_index = average_index(base)
                # current_index = average_index(current)
                # err = abs(1 - (base_index - current_index))
                # print(base_index, current_index, err)

                # base_index = average_index(cumulative_probability)
                # new_index = average_index(current)
                # if isnan(new_index):
                #     continue
                # np.divide(current, np.sum(current), current)
                # np.subtract(1, current, current)
                # print(flip,p_forward,p_backward,current)
                if use_flip < 0 or delta < lowest_err:
                    use_flip = flip
                    lowest_err = score

                # cumulative_deltas[flip] += 0

                # for k in range(0, N - 1):
                #     value = current[k] * cumulative_probability[k + 1]
                #     if use_flip < 0 or value > lowest_err:
                #         use_flip = flip
                #         lowest_err = value
                # print(flip, highest_value)
            else:
                # p_next = np.zeros(N)
                # for i in range(0, N):
                #     P = 0.0
                #     for j in range(0, N):
                #         if i == j:
                #             continue
                #         P += base[i] * (1 - base[j])
                #     p_next[i] = P
                # base = p_next

                # base[0] = 0
                np.divide(base, np.sum(base), base)
                bases.append((base.copy(), last_flip))
                # bases.insert(0, base.copy())
                # cumulative_probability = compute_cumulative_probability(N, bases)
                # p_forward = 0
                # p_backward = 0
                # for i in range(1, N):
                #     p_forward += cumulative_probability[i] * base[i - 1]
                # for i in range(0, N - 1):
                #     p_backward += cumulative_probability[i] * base[i + 1]
                print('Base', base)
                # # # np.subtract(1, base, base)
                # # # print(cumulative_probability)
                # shift_left = np.roll(cumulative_probability, -1)
                # shift_left[-1] = 0.0
                # # # # print('Shift Left', p_forward, shift_left)
                # shift_right = np.roll(cumulative_probability, 1)
                # shift_right[0] = 0.0
                # # # # print('Shift Right', p_backward, shift_right)
                # p_next = np.add(np.multiply(shift_left, 0.5), np.multiply(shift_right, 0.5))
                # p_next[0] = 0
                # np.divide(p_next, np.sum(p_next), p_next)
                # # # # print('Next', p_next)
                # # # # # print(cumulative_probability)
                # # # # # print(base)
                # np.multiply(base, p_next, cumulative_probability)
                # cumulative_probability[0] = 0
                # # # # # np.multiply(cumulative_probability, shift_right, cumulative_probability)
                # np.divide(cumulative_probability, np.sum(cumulative_probability), cumulative_probability)
                cumulative_probability = compute_cumulative_probability(N, bases, flip_likelihood)
                print('Cumulative', cumulative_probability)
                print('Likelihood', flip_likelihood)

        # cumulative_probability[0] = 0
        # use_flip = -1
        # if direction < 0:
        #     use_flip = np.argmax(cumulative_deltas)
        #     if cumulative_deltas[use_flip] < 0:
        #         use_flip = np.argmin(cumulative_deltas)
        #         direction = 1
        #         # cumulative_deltas.fill(0)
        # else:
        #     use_flip = np.argmin(cumulative_deltas)
        #     if cumulative_deltas[use_flip] > 0:
        #         use_flip = np.argmax(cumulative_deltas)
        #         direction = -1
        #         # cumulative_deltas.fill(0)
        # if direction < 0:
        #     cumulative_probability[0] = 0
        # else:
        #     cumulative_probability[-1] = 0
        # np.divide(cumulative_probability, np.sum(cumulative_probability), cumulative_probability)
        # print(cumulative_deltas)

        # use_flip = -1
        # highest_p = 0
        # for i in range(0, N):
        #     p = flip_likelihood[i]
        #     if i in flips:
        #         p = -p
        #     if use_flip < 0 or p > highest_p:
        #         use_flip = i
        #         highest_p = p
        # if not use_flip in flips and highest_p < 0 or use_flip in flips and highest_p > 0:
        #     flip_likelihood[use_flip] *= -1.0

        if use_flip < 0:
            return
        last_flip = use_flip
        if use_flip in flips:
            flips.remove(use_flip)
        else:
            flips.add(use_flip)
        print('Flip', use_flip, lowest_err)
        print(flips)
        cumulative_deltas[use_flip] = -cumulative_deltas[use_flip]
        for p in samples:
            if p.x[use_flip]:
                p.y ^= 1

if __name__ == "__main__":
    main()